#[repr(transparent)]pub struct SyncView<T: ?Sized> {
inner: T,
}exclusive_wrapper #98407)Expand description
SyncView provides mutable access, also referred to as exclusive
access to the underlying value. However, it only permits immutable, or shared
access to the underlying value when that value is Sync.
While this may seem not very useful, it allows SyncView to unconditionally
implement Sync. Indeed, the safety requirements of Sync state that for SyncView
to be Sync, it must be sound to share across threads, that is, it must be sound
for &SyncView to cross thread boundaries. By design, a &SyncView<T> for non-Sync
T has no API whatsoever, making it useless, thus harmless, thus memory safe.
Certain constructs like Futures can only be used with exclusive access,
and are often Send but not Sync, so SyncView can be used as hint to the
Rust compiler that something is Sync in practice.
Examples
Using a non-Sync future prevents the wrapping struct from being Sync:
use core::cell::Cell;
async fn other() {}
fn assert_sync<T: Sync>(t: T) {}
struct State<F> {
future: F
}
assert_sync(State {
future: async {
let cell = Cell::new(1);
let cell_ref = &cell;
other().await;
let value = cell_ref.get();
}
});SyncView ensures the struct is Sync without stripping the future of its
functionality:
#![feature(exclusive_wrapper)]
use core::cell::Cell;
use core::sync::SyncView;
async fn other() {}
fn assert_sync<T: Sync>(t: T) {}
struct State<F> {
future: SyncView<F>
}
assert_sync(State {
future: SyncView::new(async {
let cell = Cell::new(1);
let cell_ref = &cell;
other().await;
let value = cell_ref.get();
})
});Parallels with a mutex
In some sense, SyncView can be thought of as a compile-time version of
a mutex, as the borrow-checker guarantees that only one &mut can exist
for any value. This is a parallel with the fact that
& and &mut references together can be thought of as a compile-time
version of a read-write lock.
Fields
inner: Texclusive_wrapper #98407)Implementations
Sourceimpl<T: ?Sized> SyncView<T>
impl<T: ?Sized> SyncView<T>
Sourcepub const fn as_pin_mut(self: Pin<&mut Self>) -> Pin<&mut T>
🔬This is a nightly-only experimental API. (exclusive_wrapper #98407)
pub const fn as_pin_mut(self: Pin<&mut Self>) -> Pin<&mut T>
exclusive_wrapper #98407)Gets pinned exclusive access to the underlying value.
SyncView is considered to structurally pin the underlying
value, which means unpinned SyncViews can produce unpinned
access to the underlying value, but pinned SyncViews only
produce pinned access to the underlying value.
Sourcepub const fn from_mut(r: &mut T) -> &mut SyncView<T> ⓘ
🔬This is a nightly-only experimental API. (exclusive_wrapper #98407)
pub const fn from_mut(r: &mut T) -> &mut SyncView<T> ⓘ
exclusive_wrapper #98407)Build a mutable reference to an SyncView<T> from
a mutable reference to a T. This allows you to skip
building an SyncView with SyncView::new.
Sourcepub const fn from_pin_mut(r: Pin<&mut T>) -> Pin<&mut SyncView<T>>
🔬This is a nightly-only experimental API. (exclusive_wrapper #98407)
pub const fn from_pin_mut(r: Pin<&mut T>) -> Pin<&mut SyncView<T>>
exclusive_wrapper #98407)Build a pinned mutable reference to an SyncView<T> from
a pinned mutable reference to a T. This allows you to skip
building an SyncView with SyncView::new.
Sourceimpl<T: ?Sized + Sync> SyncView<T>
impl<T: ?Sized + Sync> SyncView<T>
Sourcepub const fn as_pin(self: Pin<&Self>) -> Pin<&T>
🔬This is a nightly-only experimental API. (exclusive_wrapper #98407)
pub const fn as_pin(self: Pin<&Self>) -> Pin<&T>
exclusive_wrapper #98407)Gets pinned shared access to the underlying value.
SyncView is considered to structurally pin the underlying
value, which means unpinned SyncViews can produce unpinned
access to the underlying value, but pinned SyncViews only
produce pinned access to the underlying value.